@@ -1,6 +1,6 @@
 from atexit import register as at_exit
 from hashlib import md5
-from os import getenv, getpid, makedirs, mkdir, remove, rmdir, setpgrp
+from os import environ, getenv, getpid, makedirs, mkdir, remove, rmdir, setpgrp
 from os.path import isfile, join
 from shlex import quote
 from shutil import rmtree
@@ -10,6 +10,7 @@ from time import sleep
 
 from bundlewrap.exceptions import BundleError, RepositoryError
 from bundlewrap.items import Item
+from bundlewrap.operations import RunResult
 from bundlewrap.utils import cached_property
 from bundlewrap.utils.text import is_subdirectory, mark_for_translation as _, randstr
 from bundlewrap.utils.ui import io
@@ -30,86 +31,6 @@ def is_ref(rev):
     return False
 
 
-def clone_to_dir(remote_url, rev):
-    """
-    Clones the given URL to a temporary directory, using a shallow clone
-    if the given revision is definitely not a commit hash. Clones to
-    the base directory $BW_GIT_DEPLOY_CACHE if set.
-
-    Returns the path to the repo directory and to another directory to
-    be deleted when the process exits (may be None).
-    """
-    repo_dir_hashed = md5(remote_url.encode('UTF-8')).hexdigest()
-
-    cache_dir_env = getenv("BW_GIT_DEPLOY_CACHE")
-    if cache_dir_env:
-        # Do not remove this, because it was not created by us.
-        remove_dir = None
-        repo_dir = join(cache_dir_env, repo_dir_hashed)
-        lock_dir = join(cache_dir_env, repo_dir_hashed + ".bw_lock")
-    else:
-        remove_dir = join(gettempdir(), "bw-git-cache-{}".format(getpid()))
-        repo_dir = join(remove_dir, repo_dir_hashed)
-        lock_dir = join(remove_dir, repo_dir_hashed + ".bw_lock")
-
-    makedirs(repo_dir, exist_ok=True)
-
-    io.debug(_("{pid}: lock_dir {lock_dir}").format(lock_dir=lock_dir, pid=getpid()))
-    io.debug(_("{pid}: remove_dir {remove_dir}").format(remove_dir=remove_dir, pid=getpid()))
-    io.debug(_("{pid}: repo_dir {repo_dir}").format(repo_dir=repo_dir, pid=getpid()))
-
-    if is_ref(rev) and not remote_url.startswith('http'):
-        git_cmdline = ["clone", "--bare", "--depth", "1", "--no-single-branch", remote_url, "."]
-    else:
-        git_cmdline = ["clone", "--bare", remote_url, "."]
-
-    # Use a lock directory to cooperate with other running instances
-    # of bw (in cases where $BW_GIT_DEPLOY_CACHE is used).
-    while True:
-        try:
-            mkdir(lock_dir)
-            io.debug(_("{pid}: Have lock on {lock_dir}").format(
-                lock_dir=lock_dir,
-                pid=getpid(),
-            ))
-            break
-        except FileExistsError:
-            io.debug(_("{pid}: Waiting for lock on {lock_dir} ...").format(
-                lock_dir=lock_dir,
-                pid=getpid(),
-            ))
-            sleep(1)
-
-    try:
-        # We now have a lock, but another process may have cloned
-        # the repo in the meantime. (It is vital to use a git command
-        # here, which does not traverse to parent directories.)
-        try:
-            git_command(
-                ["rev-parse", "--resolve-git-dir", "."],
-                repo_dir,
-                error_messages=False,
-            )
-            io.debug(_("{pid}: Repo already existed in {repo_dir}").format(
-                repo_dir=repo_dir,
-                pid=getpid(),
-            ))
-        except RuntimeError:
-            git_command(git_cmdline, repo_dir)
-            io.debug(_("{pid}: Cloned repo to {repo_dir}").format(
-                repo_dir=repo_dir,
-                pid=getpid(),
-            ))
-    finally:
-        rmdir(lock_dir)
-        io.debug(_("{pid}: Released lock on {lock_dir}").format(
-            lock_dir=lock_dir,
-            pid=getpid(),
-        ))
-
-    return repo_dir, remove_dir
-
-
 def get_local_repo_path(bw_repo_path, repo_name):
     """
     From the given BundleWrap repo, get the filesystem path to the git
@@ -147,39 +68,6 @@ def get_local_repo_path(bw_repo_path, repo_name):
     ))
 
 
-def git_command(cmdline, repo_dir, error_messages=True):
-    """
-    Runs the given git command line in the given directory.
-
-    Returns stdout of the command.
-    """
-    cmdline = ["git"] + cmdline
-    io.debug(_("running '{}' in {}").format(
-        " ".join(cmdline),
-        repo_dir,
-    ))
-    git_process = Popen(
-        cmdline,
-        cwd=repo_dir,
-        preexec_fn=setpgrp,
-        stderr=PIPE,
-        stdout=PIPE,
-    )
-    stdout, stderr = git_process.communicate()
-    # FIXME integrate this into Item._command_results
-    if git_process.returncode != 0:
-        if error_messages:
-            io.stderr(_("{} failed command: {}").format(getpid(), " ".join(cmdline)))
-            io.stderr(_("{} failed in dir: {}").format(getpid(), repo_dir))
-            io.stderr(_("{} stdout:\n{}").format(getpid(), stdout))
-            io.stderr(_("{} stderr:\n{}").format(getpid(), stderr))
-        raise RuntimeError(_("`git {command}` failed in {dir}").format(
-            command=cmdline[1],
-            dir=repo_dir,
-        ))
-    return stdout.decode('utf-8').strip()
-
-
 class GitDeploy(Item):
     """
     Facilitates deployment of a given rev from a local git repo to a
@@ -204,7 +92,7 @@ class GitDeploy(Item):
     @cached_property
     def _expanded_rev(self):
         git_cmdline = ["rev-parse", self.attributes['rev']]
-        return git_command(
+        return self.run_git(
             git_cmdline,
             self._repo_dir,
         )
@@ -212,7 +100,10 @@ class GitDeploy(Item):
     @cached_property
     def _repo_dir(self):
         if "://" in self.attributes['repo']:
-            repo_dir, remove_dir = clone_to_dir(self.attributes['repo'], self.attributes['rev'])
+            repo_dir, remove_dir = self.clone_to_dir(
+                self.attributes['repo'],
+                self.attributes['rev'],
+            )
             if remove_dir is not None:
                 io.debug(_("registering {} for deletion on exit").format(remove_dir))
                 at_exit(rmtree, remove_dir, ignore_errors=True)
@@ -260,7 +151,7 @@ class GitDeploy(Item):
         archive_local = NamedTemporaryFile(delete=False)
         try:
             archive_local.close()
-            git_command(
+            self.run_git(
                 ["archive", "-o", archive_local.name, self._expanded_rev],
                 self._repo_dir,
             )
@@ -307,4 +198,127 @@ class GitDeploy(Item):
         else:
             return {'rev': status_result.stdout.decode('utf-8').strip()}
 
+    def run_git(self, cmdline, repo_dir):
+        """
+        Runs the given git command line in the given directory.
+
+        Returns stdout of the command.
+        """
+        git_env = environ.copy()
+        git_env['GIT_TERMINAL_PROMPT'] = '0'
+        cmdline = ["git"] + cmdline
+        io.debug(_("running '{}' in {}").format(
+            " ".join(cmdline),
+            repo_dir,
+        ))
+        git_process = Popen(
+            cmdline,
+            cwd=repo_dir,
+            env=git_env,
+            preexec_fn=setpgrp,
+            stderr=PIPE,
+            stdout=PIPE,
+        )
+        stdout, stderr = git_process.communicate()
+        result = RunResult()
+        result.stdout = stdout
+        result.stderr = stderr
+        result.return_code = git_process.returncode
+        self._command_results.append({
+            'command': " ".join(cmdline),
+            'result': result,
+        })
+
+        if result.return_code != 0:
+            raise RuntimeError(_(
+                "`git {command}` failed in {dir} for {item} from bundle {bundle}:\n"
+                "{stdout}\n{stderr}"
+            ).format(
+                command=" ".join(cmdline[1:]),
+                dir=repo_dir,
+                item=self.id,
+                bundle=self.bundle.name,
+                stdout=result.stdout.decode('utf-8'),
+                stderr=result.stderr.decode('utf-8'),
+            ))
+        return stdout.decode('utf-8').strip()
+
+    def clone_to_dir(self, remote_url, rev):
+        """
+        Clones the given URL to a temporary directory, using a shallow clone
+        if the given revision is definitely not a commit hash. Clones to
+        the base directory $BW_GIT_DEPLOY_CACHE if set.
+
+        Returns the path to the repo directory and to another directory to
+        be deleted when the process exits (may be None).
+        """
+        repo_dir_hashed = md5(remote_url.encode('UTF-8')).hexdigest()
+
+        cache_dir_env = getenv("BW_GIT_DEPLOY_CACHE")
+        if cache_dir_env:
+            # Do not remove this, because it was not created by us.
+            remove_dir = None
+            repo_dir = join(cache_dir_env, repo_dir_hashed)
+            lock_dir = join(cache_dir_env, repo_dir_hashed + ".bw_lock")
+        else:
+            remove_dir = join(gettempdir(), "bw-git-cache-{}".format(getpid()))
+            repo_dir = join(remove_dir, repo_dir_hashed)
+            lock_dir = join(remove_dir, repo_dir_hashed + ".bw_lock")
+
+        makedirs(repo_dir, exist_ok=True)
+
+        io.debug(_("{pid}: lock_dir {lock_dir}").format(lock_dir=lock_dir, pid=getpid()))
+        io.debug(_("{pid}: remove_dir {remove_dir}").format(remove_dir=remove_dir, pid=getpid()))
+        io.debug(_("{pid}: repo_dir {repo_dir}").format(repo_dir=repo_dir, pid=getpid()))
+
+        if is_ref(rev) and not remote_url.startswith('http'):
+            git_cmdline = ["clone", "--bare", "--depth", "1", "--no-single-branch", remote_url, "."]
+        else:
+            git_cmdline = ["clone", "--bare", remote_url, "."]
+
+        # Use a lock directory to cooperate with other running instances
+        # of bw (in cases where $BW_GIT_DEPLOY_CACHE is used).
+        while True:
+            try:
+                mkdir(lock_dir)
+                io.debug(_("{pid}: Have lock on {lock_dir}").format(
+                    lock_dir=lock_dir,
+                    pid=getpid(),
+                ))
+                break
+            except FileExistsError:
+                io.debug(_("{pid}: Waiting for lock on {lock_dir} ...").format(
+                    lock_dir=lock_dir,
+                    pid=getpid(),
+                ))
+                sleep(1)
+
+        try:
+            # We now have a lock, but another process may have cloned
+            # the repo in the meantime. (It is vital to use a git command
+            # here which does not traverse to parent directories.)
+            try:
+                self.run_git(
+                    ["rev-parse", "--resolve-git-dir", "."],
+                    repo_dir,
+                )
+                io.debug(_("{pid}: Repo already existed in {repo_dir}").format(
+                    repo_dir=repo_dir,
+                    pid=getpid(),
+                ))
+            except RuntimeError:
+                self.run_git(git_cmdline, repo_dir)
+                io.debug(_("{pid}: Cloned repo to {repo_dir}").format(
+                    repo_dir=repo_dir,
+                    pid=getpid(),
+                ))
+        finally:
+            rmdir(lock_dir)
+            io.debug(_("{pid}: Released lock on {lock_dir}").format(
+                lock_dir=lock_dir,
+                pid=getpid(),
+            ))
+
+        return repo_dir, remove_dir
+
 # FIXME get_auto_deps for dir and ensure dir does not use purge
