@@ -10,10 +10,11 @@ from tomlkit import dumps as toml_dump, parse as toml_parse
 from . import operations
 from .bundle import Bundle
 from .concurrency import WorkerPool
-from .deps import find_item
+from .deps import find_item, ItemDependencyLoop
 from .exceptions import (
+    BundleError,
     GracefulApplyException,
-    ItemDependencyLoop,
+    ItemSkipped,
     NodeLockedException,
     NoSuchBundle,
     RemoteException,
@@ -27,8 +28,9 @@ from .lock import NodeLock
 from .metadata import hash_metadata
 from .utils import cached_property, error_context, get_file_contents, names
 from .utils.dicts import (
+    dict_to_text,
     dict_to_toml,
-    diff_value,
+    diff_dict,
     hash_statedict,
     set_key_at_path,
     validate_dict,
@@ -42,6 +44,7 @@ from .utils.text import (
     format_duration,
     green,
     mark_for_translation as _,
+    prefix_lines,
     red,
     toml_clean,
     validate_name,
@@ -248,10 +251,40 @@ def apply_items(
         io.progress_advance()
         results.append((item.id, status_code, duration))
 
+    def handle_exception(task_id, exc, traceback):
+        item_id = task_id.split(":", 1)[1]
+        item = find_item(item_id, item_queue.pending_items)
+
+        for skipped_item in item_queue.item_failed(item):
+            handle_apply_result(
+                node,
+                skipped_item,
+                Item.STATUS_SKIPPED,
+                interactive=interactive,
+                details=Item.SKIP_REASON_DEP_FAILED,
+            )
+            results.append((skipped_item.id, Item.STATUS_SKIPPED, timedelta(0)))
+
+        handle_apply_result(
+            node,
+            item,
+            Item.STATUS_FAILED,
+            interactive=interactive,
+            details=None,
+            show_diff=False,
+        )
+        results.append((item.id, Item.STATUS_FAILED, timedelta(0)))
+        output = prefix_lines(traceback, red("│ "))
+        output += prefix_lines(repr(exc), red("│ "))
+        output += red("╵")
+        io.stderr(output)
+        io.progress_advance()
+
     worker_pool = WorkerPool(
         tasks_available,
         next_task,
         handle_result=handle_result,
+        handle_exception=handle_exception,
         pool_id="apply_{}".format(node.name),
         workers=workers,
     )
@@ -277,7 +310,7 @@ def _flatten_group_hierarchy(groups):
 
     # dict mapping groups to parent groups
     parent_groups = {}
-    for child_group in child_groups.keys():
+    for child_group in child_groups:
         parent_groups[child_group] = []
         for parent_group, subgroups in child_groups.items():
             if child_group in subgroups:
@@ -302,7 +335,7 @@ def _flatten_group_hierarchy(groups):
                 break
         order.append(top_level_group)
         del parent_groups[top_level_group]
-        for group in parent_groups.keys():
+        for group in parent_groups:
             if top_level_group in parent_groups[group]:
                 parent_groups[group].remove(top_level_group)
 
@@ -312,32 +345,31 @@ def _flatten_group_hierarchy(groups):
 def format_item_command_results(results):
     output = ""
 
-    for i in range(len(results)):
-        stdout = results[i]['result'].stdout_text.strip()
-        stderr = results[i]['result'].stderr_text.strip()
+    for result in results:
+        stdout = result['result'].stdout_text.strip()
+        stderr = result['result'].stderr_text.strip()
 
         # show command
-        output += "\n{b}".format(b=red('│'))
-        output += "\n{b} {command} (return code: {code}{no_output})".format(
+        output += "{b}".format(b=red('│'))
+        output += "\n{b} {command} (return code: {code}{no_output})\n".format(
             b=red('├─'),
-            command=bold(results[i]['command']),
-            code=bold(results[i]['result'].return_code),
+            command=bold(result['command']),
+            code=bold(result['result'].return_code),
             no_output='' if stdout or stderr else '; no output'
         )
 
         # show output
-        lines = []
+        lines = ""
         if stdout or stderr:
-            output += "\n{b}".format(b=red("│ "))
+            lines += "\n"
             if stdout:
-                lines += stdout.strip().split('\n')
+                lines += stdout.strip()
             if stderr:
-                lines += stderr.strip().split('\n')
+                lines += stderr.strip()
 
-        for k in range(len(lines)):
-            output += "\n{b} {line}".format(b=red("│ "), line=lines[k])
+        output += prefix_lines(lines, red("│ "))
 
-    output += red("\n╵ ")
+    output += red("╵")
     return output.lstrip('\n')
 
 
@@ -373,14 +405,8 @@ def format_item_result(
                 status=status,
                 x=bold(red("✘")),
             )
-            diff = "\n"
-            for key in sorted(details[2]):
-                diff += diff_value(key, details[1][key], details[0][key]) + "\n"
-            for line in diff.splitlines():
-                output += "{x} {line}\n".format(
-                    line=line,
-                    x=red("│"),
-                )
+            diff = diff_dict(details[1], details[0], skip_missing_in_target=True)
+            output += prefix_lines(diff, f"{red('│')} ")
             output += red("╵")
             return output
         else:
@@ -425,17 +451,13 @@ def format_item_result(
                 status=status,
             )
             diff = "\n"
-            if created or deleted:
-                for key, value in sorted(details.items()):
-                    diff += f"{bold(key)}  {value}\n"
+            if created:
+                diff += dict_to_text(details, value_color=green)
+            elif deleted:
+                diff += dict_to_text(details, value_color=red)
             else:
-                for key in sorted(details[2]):
-                    diff += diff_value(key, details[1][key], details[0][key]) + "\n"
-            for line in diff.splitlines():
-                output += "{x} {line}\n".format(
-                    line=line,
-                    x=green("│"),
-                )
+                diff += diff_dict(details[1], details[0], skip_missing_in_target=True)
+            output += prefix_lines(diff, f"{green('│')} ")
             output += green("╵")
             return output
         else:
@@ -480,7 +502,7 @@ class Node:
         OS_FAMILY_REDHAT
 
     OS_FAMILY_UNIX = OS_FAMILY_BSD + OS_FAMILY_LINUX
-    OS_KNOWN = OS_FAMILY_UNIX + ('kubernetes',)
+    OS_KNOWN = OS_FAMILY_UNIX + ('kubernetes', 'routeros')
 
     def __init__(self, name, attributes=None):
         if attributes is None:
@@ -616,10 +638,21 @@ class Node:
 
     @cached_property
     def items(self):
+        items = {}
         if not self.dummy:
             for bundle in self.bundles:
                 for item in bundle.items:
-                    yield item
+                    if item.id in items:
+                        raise BundleError(_(
+                            "duplicate definition of {item} in bundles '{bundle1}' and '{bundle2}'"
+                        ).format(
+                            item=item.id,
+                            bundle1=item.bundle.name,
+                            bundle2=items[item.id].bundle.name,
+                        ))
+                    else:
+                        items[item.id] = item
+        return set(items.values())
 
     @cached_property
     def magic_number(self):
@@ -627,12 +660,12 @@ class Node:
 
     def apply(
         self,
-        autoskip_selector="",
-        autoonly_selector="",
+        autoskip_selector=(),
+        autoonly_selector=(),
         interactive=False,
         force=False,
         show_diff=True,
-        skip_list=tuple(),
+        skip_list=(),
         workers=4,
     ):
         if not list(self.items):
@@ -751,6 +784,7 @@ class Node:
             remote_path,
             local_path,
             add_host_keys=self._add_host_keys,
+            username=self.username,
             wrapper_inner=self.cmd_wrapper_inner,
             wrapper_outer=self.cmd_wrapper_outer,
         )
@@ -797,7 +831,7 @@ class Node:
         """
         return self.metadata
 
-    def run(self, command, data_stdin=None, may_fail=False, log_output=False):
+    def run(self, command, data_stdin=None, may_fail=False, log_output=False, user="root"):
         assert self.os in self.OS_FAMILY_UNIX
 
         if log_output:
@@ -837,8 +871,10 @@ class Node:
             data_stdin=data_stdin,
             ignore_failure=may_fail,
             log_function=log_function,
+            username=self.username,
             wrapper_inner=self.cmd_wrapper_inner,
             wrapper_outer=self.cmd_wrapper_outer,
+            user=user,
         )
 
     @cached_property
@@ -876,18 +912,33 @@ class Node:
             mode=mode,
             owner=owner,
             ignore_failure=may_fail,
+            username=self.username,
             wrapper_inner=self.cmd_wrapper_inner,
             wrapper_outer=self.cmd_wrapper_outer,
         )
 
-    def verify(self, show_all=False, show_diff=True, workers=4):
+    def verify(
+        self,
+        autoskip_selector=(),
+        autoonly_selector=(),
+        show_all=False,
+        show_diff=True,
+        workers=4,
+    ):
         result = []
         start = datetime.now()
 
         if not self.items:
             io.stdout(_("{x} {node}  has no items").format(node=bold(self.name), x=yellow("!")))
         else:
-            result = verify_items(self, show_all=show_all, show_diff=show_diff, workers=workers)
+            result = verify_items(
+                self,
+                autoskip_selector=autoskip_selector,
+                autoonly_selector=autoonly_selector,
+                show_all=show_all,
+                show_diff=show_diff,
+                workers=workers,
+            )
 
         return {
             'good': result.count(True),
@@ -932,7 +983,14 @@ for attr, default in GROUP_ATTR_DEFAULTS.items():
     setattr(Node, attr, build_attr_property(attr, default))
 
 
-def verify_items(node, show_all=False, show_diff=True, workers=1):
+def verify_items(
+    node,
+    autoskip_selector=(),
+    autoonly_selector=(),
+    show_all=False,
+    show_diff=True,
+    workers=1,
+):
     items = []
     for item in node.items:
         if not item.triggered:
@@ -979,33 +1037,37 @@ def verify_items(node, show_all=False, show_diff=True, workers=1):
                 return {
                     'task_id': node.name + ":" + item.bundle.name + ":" + item.id,
                     'target': item.verify,
+                    'kwargs': {
+                        'autoskip_selector': autoskip_selector,
+                        'autoonly_selector': autoonly_selector,
+                    },
                 }
 
     def handle_exception(task_id, exception, traceback):
         node_name, bundle_name, item_id = task_id.split(":", 2)
         io.progress_advance()
-        if isinstance(exception, NotImplementedError):
-            io.stdout(_("{x} {node}  {bundle}  {item}  (does not support verify)").format(
-                bundle=bold(bundle_name),
-                item=item_id,
-                node=bold(node_name),
-                x=cyan("?"),
-            ))
+        if isinstance(exception, (ItemSkipped, NotImplementedError)):
+            pass
         else:
             # Unlike with `bw apply`, it is OK for `bw verify` to encounter
             # exceptions when getting an item's status. `bw verify` doesn't
             # care about dependencies and therefore cannot know that looking
             # up a database user requires the database to be installed in
             # the first place.
-            io.debug("exception while verifying {}:".format(task_id))
-            io.debug(traceback)
-            io.debug(repr(exception))
-            io.stdout(_("{x} {node}  {bundle}  {item}  (unable to get status, check --debug for details)").format(
+            item = node.get_item(task_id.split(":", 2)[2])
+            output = _("{x} {node}  {bundle}  {item}  (unable to get status)\n").format(
                 bundle=bold(bundle_name),
                 item=item_id,
                 node=bold(node_name),
-                x=cyan("?"),
-            ))
+                x=bold(cyan("?")),
+            )
+            output += prefix_lines(traceback, cyan("│ "))
+            output += prefix_lines(repr(exception), cyan("│ "))
+            if item._command_results:
+                output += format_item_command_results(item._command_results)
+                # free up memory
+                del item._command_results
+            io.stderr(output)
         return None  # count this result as "unknown"
 
     def handle_result(task_id, return_value, duration):
@@ -1023,12 +1085,12 @@ def verify_items(node, show_all=False, show_diff=True, workers=1):
                 details_text = ", ".join(sorted(display[2]))
             if show_diff:
                 diff = "\n"
-                if item_status.must_be_created or item_status.must_be_deleted:
-                    for key, value in sorted(display.items()):
-                        diff += f"{bold(key)}  {value}\n"
+                if item_status.must_be_created:
+                    diff += dict_to_text(display, value_color=green)
+                elif item_status.must_be_deleted:
+                    diff += dict_to_text(display, value_color=red)
                 else:
-                    for key in sorted(display[2]):
-                        diff += diff_value(key, display[1][key], display[0][key]) + "\n"
+                    diff += diff_dict(display[1], display[0], skip_missing_in_target=True)
                 output = "{x} {node}  {bundle}  {item}  {details}\n".format(
                     bundle=bold(bundle_name),
                     details=details_text,
@@ -1036,11 +1098,7 @@ def verify_items(node, show_all=False, show_diff=True, workers=1):
                     node=bold(node_name),
                     x=red("✘"),
                 )
-                for line in diff.splitlines():
-                    output += "{x} {line}\n".format(
-                        line=line,
-                        x=red("│"),
-                    )
+                output += prefix_lines(diff, red('│ '))
                 io.stderr(output + red("╵"))
             else:
                 io.stderr("{x} {node}  {bundle}  {item}  {details}".format(
